#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
#  SPDX-License-Identifier: Apache-2.0
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import getopt
import logging
import os
import pathlib
import sys
import signal
import threading
from logging.handlers import RotatingFileHandler
from os.path import expanduser

from c8ylp.rest_client.c8yclient import CumulocityClient
from c8ylp.tcp_socket.tcp_server import TCPServer
from c8ylp.websocket_client.ws_client import WebsocketClient

PIDFILE = '/var/run/c8ylp/c8ylp'


def start():
    home = expanduser('~')
    path = pathlib.Path(home + '/.c8ylp')
    loglevel = logging.INFO
    path.mkdir(parents=True, exist_ok=True)
    logger = logging.getLogger()
    logger.setLevel(loglevel)
    log_file_formatter = logging.Formatter(
        '%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s')
    log_console_formatter = logging.Formatter('%(message)s')

    # Set default log format
    if len(logger.handlers) == 0:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_console_formatter)
        console_handler.setLevel(loglevel)
        logger.addHandler(console_handler)
    else:
        handler = logger.handlers[0]
        handler.setFormatter(log_console_formatter)

    # Max 5 log files each 10 MB.
    rotate_handler = RotatingFileHandler(filename=path / 'localproxy.log', maxBytes=10000000,
                                         backupCount=5)
    rotate_handler.setFormatter(log_file_formatter)
    rotate_handler.setLevel(loglevel)
    # Log to Rotating File
    logger.addHandler(rotate_handler)
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h:d:c:t:u:p:kvs",
                                   ["help", "hostname=", "device=", "extype=", "config=", "tenant=", "username=",
                                    "password=", "tfacode=", "port=", "kill", "tcpsize=", "tcptimeout=", "verbose",
                                    "scriptmode",
                                    "ignore-ssl-validate"])
    except getopt.GetoptError as e:
        logging.error(e)
        help()

    logging.debug(f'OPTIONS: {opts}')
    # if len(opts) == 0:
    #    print(help())
    host = os.environ.get('C8Y_HOST')
    device = os.environ.get('C8Y_DEVICE')

    extype = os.environ.get('C8Y_EXTYPE') if os.environ.get('C8Y_EXTYPE') is not None else 'c8y_Serial'
    config_name = os.environ.get('C8Y_CONFIG') if os.environ.get('C8Y_CONFIG') is not None else 'Passthrough'
    tenant = os.environ.get('C8Y_TENANT')
    user = os.environ.get('C8Y_USER')
    password = os.environ.get('C8Y_PASSWORD')
    tcp_size = int(os.environ.get('C8Y_TCPSIZE')) if os.environ.get('C8Y_TCPSIZE') is not None else 32768
    tcp_timeout = int(os.environ.get('C8Y_TCPTIMEOUT')) if os.environ.get('C8Y_TCPTIMEOUT') is not None else 0
    port = int(os.environ.get('C8Y_PORT')) if os.environ.get('C8Y_PORT') is not None else 2222
    token = os.environ.get('C8Y_TOKEN')
    tfacode = None
    script_mode = False
    ignore_ssl_validate = False
    for option_key, option_value in opts:
        if option_key in ('-h', '--hostname'):
            host = option_value
        elif option_key in ('-d', '--device'):
            device = option_value
        elif option_key in '--extype':
            extype = option_value
        elif option_key in ('-c', '--config'):
            config_name = option_value
        elif option_key in ('-t', '--tenant'):
            tenant = option_value
        elif option_key in ('-u', '--username'):
            user = option_value
        elif option_key in ('-p', '--password'):
            password = option_value
        elif option_key in ['--tfacode']:
            tfacode = option_value
        elif option_key in ['--port']:
            port = int(option_value)
        elif option_key in ['-k', '--kill']:
            kill_existing_instances()
        elif option_key in ['--tcpsize']:
            tcp_size = int(option_value)
        elif option_key in ['--tcptimeout']:
            tcp_timeout = int(option_value)
        elif option_key in ['-v', '--verbose']:
            verbose_log()
        elif option_key in ['-s', '--scriptmode']:
            script_mode = True
        elif option_key in ['--ignore-ssl-validate']:
            ignore_ssl_validate = True
        elif option_key in ['--help']:
            help()
    validate_parameter(host, device, extype, config_name,
                       tenant, user, password, token)
    upsert_pid_file(device, host, config_name, user)
    client = CumulocityClient(host, tenant, user, password, tfacode, ignore_ssl_validate)
    tenant_id_valid = client.validate_tenant_id()
    if tenant_id_valid is not None:
        logging.warn(f'WARNING: Tenant ID {tenant} does not exist. Try using this Tenant ID {tenant_id_valid} next time!')
    session = None
    if token:
        client.validate_token()
    else:
        session = client.retrieve_token()
    mor = client.read_mo(device, extype)
    config_id = client.get_config_id(mor, config_name)
    device_id = client.get_device_id(mor)
    
    is_authorized = client.validate_remote_access_role()
    if not is_authorized:
        logging.error(f'User {user} is not authorized to use Cloud Remote Access. Contact your Cumulocity Admin!')
        sys.exit(1)
    websocket_client = WebsocketClient(
        host, tenant, user, password, config_id, device_id, session, token, ignore_ssl_validate)
    wst = websocket_client.connect()
    tcp_server = TCPServer(port, websocket_client, tcp_size, tcp_timeout, wst, script_mode)
    # TCP is blocking...
    websocket_client.tcp_server = tcp_server
    try:
        tcp_server.start()
    except Exception as ex:
        logging.error(f'Error on TCP-Server {ex}')
    finally:
        clean_pid_file(None)
        tcp_server.stop()
        sys.exit(0)


def verbose_log():
    logging.info(f'Verbose logging activated.')
    logging.getLogger().setLevel(logging.DEBUG)
    for handler in logging.getLogger().handlers:
        handler.setLevel(logging.DEBUG)


def upsert_pid_file(device, url, config, user):
    try:
        clean_pid_file(None)
        pid_file_text = get_pid_file_text(device, url, config, user)
        logging.debug(f'Adding {pid_file_text} to PID-File {PIDFILE}')
        if not os.path.exists(PIDFILE):
            if not os.path.exists(os.path.dirname(PIDFILE)):
                os.makedirs(os.path.dirname(PIDFILE))
            file = open(PIDFILE, 'w')
            file.seek(0)
        else:
            file = open(PIDFILE, 'a+')
            file.seek(0)
        file.write(pid_file_text)
        file.write('\n')
    except PermissionError:
        logging.error(
            f'Could not write PID-File {PIDFILE}. Please create the folder manually and assign the correct permissions.')
        sys.exit(1)


def get_pid_file_text(device, url, config, user):
    pid = str(os.getpid())
    return f'{pid},{url},{device},{config},{user}'


def get_pid_from_line(line):
    return int(str.split(line, ',')[0])


def pid_is_active(pid):
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def clean_pid_file(pid):
    if os.path.exists(PIDFILE):
        logging.debug(f'Cleaning Up PID {pid} in PID-File {PIDFILE}')
        pid = pid if pid is not None else os.getpid()
        with open(PIDFILE, "r+") as file:
            lines = file.readlines()
            file.seek(0)
            for line in lines:
                if get_pid_from_line(line) != pid:
                    file.write(line)
            file.truncate()
            if os.path.getsize(PIDFILE) == 0:
                os.remove(PIDFILE)


def kill_existing_instances():
    if os.path.exists(PIDFILE):
        file = open(PIDFILE)
        for line in file:
            other_pid = get_pid_from_line(line)
            if pid_is_active(other_pid):
                logging.info(
                    f'Killing other running Process with PID {other_pid}')
                os.kill(get_pid_from_line(line), 9)
            clean_pid_file(other_pid)


def validate_parameter(host, device, extpye, config_name, tenant, user, password, token):
    if not host:
        logging.error(f'Hostname is missing!')
        print('Mandatory parameter -h, --hostname is missing')
        print(_help_message())
        sys.exit(1)

    if not device:
        logging.error(f'Device Name is missing!')
        print('Mandatory parameter -d, --device is missing')
        print(_help_message())
        sys.exit(1)

    if not config_name:
        logging.error(f'Configuration Name is missing!')
        print('Mandatory parameter -c, --config is missing')
        print(_help_message())
        sys.exit(1)

    if not tenant and not token:
        logging.error(f'Tenant is missing!')
        print('Mandatory parameter -t, --tenant is missing')
        print(_help_message())
        sys.exit(1)

    if not user and not token:
        logging.error(f'User is missing!')
        print('Mandatory parameter -u, --user is missing')
        print(_help_message())
        sys.exit(1)

    if not password and not token:
        logging.error(f'Password is missing!')
        print('Mandatory parameter -p, --password is missing')
        print(_help_message())
        sys.exit(1)


def help():
    print(_help_message())
    sys.exit(2)


def _help_message() -> str:
    return str('Usage: c8ylp [params]\n'
               '\n'
               'Parameter:\n'
               ' --help                 show this help message and exit\n'
               ' -h, --hostname         REQUIRED, the Cumulocity Hostname\n'
               ' -d, --device           REQUIRED, the Device Name (ext. Id of Cumulocity)\n'
               ' --extype               OPTIONAL, the external Id Type. Default: "c8y_Serial"\n'
               ' -c, --config           OPTIONAL, the name of the C8Y Remote Access Configuration. Default: "Passthrough"\n'
               ' -t, --tenant           REQUIRED, the tenant Id of Cumulocity\n'
               ' -u, --user             REQUIRED, the username of Cumulocity\n'
               ' -p, --password         REQUIRED, the password of Cumulocity\n'
               ' --tfacode              OPTIONAL, the TFA Code when an user with the Option "TFA enabled" is used\n'
               ' --port                 OPTIONAL, the TCP Port which should be opened. Default: 2222\n'
               ' -k, --kill             OPTIONAL, kills all existing processes of c8ylp\n'
               ' --tcpsize              OPTIONAL, the TCP Package Size. Default: 32768\n'
               ' --tcptimeout           OPTIONAL, Timeout in sec. for inactivity. Can be activited with values > 0. Default: Deactivated\n'
               ' -v, --verbose          OPTIONAL, Print Debug Information into the Logs and Console when set.\n'
               ' -s, --scriptmode       OPTIONAL, Stops the TCP Server after first connection. No automatical restart!\n'
               ' --ignore-ssl-validate  OPTIONAL, Ignore Validation for SSL Certificates while connecting to Websocket'
               '\n')


if __name__ == '__main__':
    start()



#  Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
#  SPDX-License-Identifier: Apache-2.0
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import socket
import sys
import threading


class TCPServer:

    def __init__(self, port, web_socket_client, tcp_buffer_size, tcp_timeout, wst, script_mode):
        self.port = port
        self.sock = None
        self.connection = None
        self.web_socket_client = web_socket_client
        self.tcp_open_event = None
        self.tcp_buffer_size = tcp_buffer_size
        self.conn_is_closed = False
        self.tcp_timeout = tcp_timeout
        self._tcp_open_event = threading.Event()
        self.wst = wst
        self.tcp_timeout_counter = None
        self.script_mode = script_mode
        self.logger = logging.getLogger(__name__)

    def start(self):
        if self.web_socket_client.is_ws_available():
            self._start_server()
            self._start_connection()

    def _start_connection(self):
        try:
            # Listen for incoming connections
            self.conn_is_closed = False
            self._tcp_timeout_counter = 0
            self._tcp_open_event = threading.Event()
            self.sock.listen(1)
            self.logger.info(f'Waiting for incoming connections...')
            self.connection, client_address = self.sock.accept()
            #if self.tcp_timeout > 0:
            self.connection.settimeout(1)
            self.logger.info(f'TCP Client connected: {client_address}')
            self._tcp_open_event.set()
            while not self.conn_is_closed and self.connection:
                try:
                    if self.tcp_timeout > 0:
                        self.logger.debug(f'Waiting for TCP-Data... Timeout-Counter: {self._tcp_timeout_counter}s')
                    if self.connection:
                        data = self.connection.recv(self.tcp_buffer_size)
                        self.logger.debug(f'TCP Data Received: {data}')
                        self._tcp_timeout_counter  = 0
                        if data:
                            if self.web_socket_client.is_ws_available():
                                self.logger.debug(f'Sent Data to WebSocket...')
                                self.web_socket_client.web_socket.sock.send_binary(
                                    data)
                        else:
                            break
                    else:
                        break
                except socket.timeout:
                    #self.web_socket_client.reconnect()
                    if self.tcp_timeout > 0 and self._tcp_timeout_counter >= self.tcp_timeout:
                        self.logger.debug(
                            f'TCP Timeout {self.tcp_timeout} reached!')
                        break
                    else:
                        self._tcp_timeout_counter += 1
                        continue

                except Exception as ex:
                    self.logger.debug(f'Type of TCP Error {type(ex)}')
                    self.logger.error(ex)
                    break
            # Restart the TCP Server
            if not self.script_mode:
                self._restart()
        except Exception as ex:
            self.logger.error(ex)

    def _start_server(self):
        self.logger.info(
            f'Starting TCP Server on localhost and port {self.port} ... ')
        # Create a TCP/IP socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_address = ('localhost', self.port)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(server_address)
        self.logger.info(
            f'TCP Server on localhost and port {self.port} successfully started.')

    def _restart(self):
        self.logger.info(
            f'Restarting TCP Connection...')
        self.stop_connection()
        #self.stop()
        if self.web_socket_client.is_ws_available():
            #self._start_server()
            self._start_connection()

    def is_tcp_socket_connected(self):
        if self.connection:
            return True
        tcp_result = self._tcp_open_event.wait()
        return tcp_result

    def is_tcp_socket_available(self):
        tcp_result = self._tcp_open_event.is_set()
        return tcp_result

    def stop_connection(self):
        # Close Client connection...
        if self.connection and not self.conn_is_closed:
            self.logger.info(
                f'Stopping TCP Connection {self.connection.getpeername()}')
            self.connection.close()
            # self.connection.shutdown(socket.SHUT_RDWR)
            self.connection = None
            self.conn_is_closed = True

    def stop(self):
        self.logger.info(f'Shutting down TCP Server...')
        self.stop_connection()
        if self.sock:
            # self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()
        self.logger.info(f'TCP Server shutdown successful!')
        # sys.exit()



#  Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
#  SPDX-License-Identifier: Apache-2.0
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import json
import os
import logging
import sys
from base64 import b64encode

import requests


class CumulocityClient:

    def __init__(self, hostname: str, tenant: str, user: str, password: str, tfacode: str,
                 ignore_ssl_validate: bool = False):
        self.hostname = hostname
        self.tenant = tenant
        self.user = user
        self.password = password
        self.tfacode = tfacode
        self.session = requests.Session()
        self.token = os.environ.get('C8Y_TOKEN')
        if hostname.startswith('http'):
            self.url = hostname
        else:
            self.url = f'https://{hostname}'
        if ignore_ssl_validate:
            self.session.verify = False
        self.logger = logging.getLogger(__name__)
    
    def validate_tenant_id(self):
        tenant_id = None
        current_user_url = self.url + f'/tenant/loginOptions'
        headers = {}
        response = self.session.get(current_user_url, headers=headers)
        self.logger.debug(f'Response received: {response}')
        if response.status_code == 200:
            login_options_body = json.loads(response.content.decode('utf-8'))
            login_options = login_options_body['loginOptions']
            for option in login_options:
                if 'initRequest' in option:
                    tenant_id = option['initRequest'].split('=')[1]
                    if self.tenant != tenant_id:
                        self.logger.debug(f'Wrong Tenant ID {self.tenant}, Correct Tenant ID: {tenant_id}')
                        self.tenant = tenant_id
                    else:
                        tenant_id = None
                    break
        
        else:
            self.logger.error(f'Error validating Tenant ID!')
        return tenant_id

    def validate_remote_access_role(self):
        is_valid = False
        current_user_url = self.url + f'/user/currentUser'
        if self.token:
            headers = {'Content-Type': 'application/json',
                       'Authorization': 'Bearer ' +self.token}
        else:
            headers = {'Content-Type': 'application/json',
                    'X-XSRF-TOKEN': self.session.cookies.get_dict()['XSRF-TOKEN']
                    }
        response = self.session.get(current_user_url, headers=headers)
        self.logger.debug(f'Response received: {response}')
        if response.status_code == 200:
            user = json.loads(response.content.decode('utf-8'))
            effective_roles = user['effectiveRoles']
            for role in effective_roles:
                if 'ROLE_REMOTE_ACCESS_ADMIN' == role['id']:
                    self.logger.debug(f'Remote Access Role assigned to User {self.user}!')
                    is_valid = True
                    break
        else:
            self.logger.error(f'Error retrieving User Data!')
            is_valid = False
        return is_valid

    def validate_token(self):
        is_valid = False
        current_user_url = self.url + f'/user/currentUser'
        headers = {'Content-Type': 'application/json', 'Authorization': 'Bearer ' + self.token}
        response = self.session.get(current_user_url, headers=headers)
        self.logger.debug(f'Response received: {response}')
        if response.status_code == 200:
            is_valid = True
        else:
            self.logger.error(f'Error validating Token {response.status_code}. Please provide Tenant User and Password!')
            del os.environ['C8Y_TOKEN']
            is_valid = False
            sys.exit(1)
        return is_valid

    def retrieve_token(self):
        oauth_url = self.url + f'/tenant/oauth?tenant_id={self.tenant}'
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = {
            'grant_type': 'PASSWORD',
            'username': self.user,
            'password': self.password,
            'tfa_code': self.tfacode
        }
        self.logger.debug(f'Sending requests to {oauth_url}')
        response = self.session.post(oauth_url,headers=headers,data=body)
        if response.status_code == 200:
            self.logger.debug(f'Authenticateion successful. Tokens have been updated {self.session.cookies.get_dict()}!')
            os.environ['C8Y_TOKEN'] = self.session.cookies.get_dict()['authorization']
        elif response.status_code == 401:
            self.logger.error(f'User {self.user} is not authorized to access Tenant {self.tenant} or TFA-Code is invalid.')
            sys.exit(1)
        else:
            self.logger.error(f'Server Error received for User {self.user} and Tenant {self.tenant}. Status Code: {response.status_code}')
            sys.exit(1)
        return self.session

    def read_ext_Id(self, device, extype):
        identiy_url = self.url + f'/identity/externalIds/{extype}/{device}'
        #auth_string = f'{self.tenant}/{self.user}:{self.password}'
        #encoded_auth_string = b64encode(
        #    bytes(auth_string, 'utf-8')).decode('ascii')
        if self.token:
            headers = {'Content-Type': 'application/json',
                       'Authorization': 'Bearer ' +self.token}
        else:
            headers = {'Content-Type': 'application/json',
                    'X-XSRF-TOKEN': self.session.cookies.get_dict()['XSRF-TOKEN']
                   #'Authorization': 'Basic ' + encoded_auth_string
                    }
        self.logger.debug(f'Sending requests to {identiy_url}')
        response = self.session.get(identiy_url, headers=headers)
        self.logger.debug(f'Response received: {response}')
        ext_id = None
        if response.status_code == 200:
            ext_id = json.loads(response.content.decode('utf-8'))
        elif response.status_code == 401:
            self.logger.error(f'User {self.user} is not authorized to read Device Data of Device {device}')
            sys.exit(1)
        elif response.status_code == 404:
            self.logger.error(f'Device {device} not found!')
            # print(f'Device {device} not found!')
            sys.exit(1)
        else:
            self.logger.error(f'Error on retrieving device. Status Code {response.status_code}')
            # print(f'Error on retrieving device. Status Code {response.status_code}')
            sys.exit(1)
        return ext_id

    def read_mo(self, device, extype):
        ext_id = self.read_ext_Id(device, extype)
        mor_id = None
        mor = None
        if ext_id['managedObject']['id']:
            mor_id = ext_id['managedObject']['id']
        if mor_id:
            managed_object_url = self.url + f'/inventory/managedObjects/{mor_id}'
            if self.token:
                headers = {'Content-Type': 'application/json',
                       'Authorization': 'Bearer ' +self.token}
            else:
                headers = {'Content-Type': 'application/json',
                    'X-XSRF-TOKEN': self.session.cookies.get_dict()['XSRF-TOKEN']
                   #'Authorization': 'Basic ' + encoded_auth_string
                    }
            #auth_string = f'{self.tenant}/{self.user}:{self.password}'
            #encoded_auth_string = b64encode(
            #    bytes(auth_string, 'utf-8')).decode('ascii')
            self.logger.debug(f'Sending requests to {managed_object_url}')
            response = self.session.get(managed_object_url, headers=headers)
            self.logger.debug(f'Response received: {response}')
            if response.status_code == 200:
                mor = json.loads(response.content.decode('utf-8'))
                return mor
            elif response.status_code == 401:
                self.logger.error(f'User {self.user} is not authorized to read Device Data of Device {device}')
                sys.exit(1)
            elif response.status_code == 404:
                self.logger.error(f'Device {device} not found!')
                sys.exit(1)
            else:
                self.logger.error(f'Error on retrieving device. Status Code {response.status_code}')
                sys.exit(1)
            return mor

    def get_config_id(self, mor, config):
        access_list = mor['c8y_RemoteAccessList']
        device = mor['name']
        config_id = None
        for remote_access in access_list:
            if not remote_access['protocol'] == 'PASSTHROUGH':
                continue
            if config and remote_access['name'] == config:
                config_id = remote_access['id']
                self.logger.info(f'Using Configuration with Name "{config}" and Remote Port {remote_access["port"]}')
                break
            if not config:
                config_id = remote_access['id']
                self.logger.info(f'Using Configuration with Name "{config}" and Remote Port {remote_access["port"]}')
                break
        if not config_id:
            if config:
                self.logger.error(
                    f'Provided config name "{config}" for "{device}" was not found or not of type "PASSTHROUGH"')
                sys.exit(1)
            else:
                self.logger.error(f'No config of Type "PASSTHROUGH" has been found for device "{device}"')
                sys.exit(1)
        return config_id

    def get_device_id(self, mor):
        return mor['id']
